@@ -23,8 +23,17 @@
 
 static const char* FILE_NAME = "Sys/IO.c";
 
+#undef MY_DIR
+#undef MY_DIRENT
+
 #if defined(_WIN32)
   #include "Sys-Windows.h"
+  
+  #define MY_DIR _WDIR
+  #define MY_DIRENT struct _wdirent
+#else
+  #define MY_DIR DIR
+  #define MY_DIRENT struct dirent
 #endif
 
 int32_t SPVM__Sys__IO__fopen(SPVM_ENV* env, SPVM_VALUE* stack) {
@@ -1368,7 +1377,17 @@ int32_t SPVM__Sys__IO__opendir(SPVM_ENV* env, SPVM_VALUE* stack) {
   }
   const char* dir = env->get_chars(env, stack, obj_dir);
   
-  DIR* dir_stream = opendir(dir);
+#if defined(_WIN32)
+  wchar_t* dir_w = utf8_to_win_wchar(env, stack, dir, &error_id, __func__, FILE_NAME, __LINE__);
+  if (error_id) {
+    return error_id;
+  }
+  
+  MY_DIR* dir_stream = _wopendir(dir_w);
+#else
+  MY_DIR* dir_stream = opendir(dir);
+#endif
+
   if (!dir_stream) {
     env->die(env, stack, "[System Error]opendir() failed:%s. $dir is \"%s\".", env->strerror_nolen(env, stack, errno), dir, __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
@@ -1390,9 +1409,14 @@ int32_t SPVM__Sys__IO__closedir(SPVM_ENV* env, SPVM_VALUE* stack) {
   if (!obj_dirp) {
     return env->die(env, stack, "The directory stream $dirp must be defined.", __func__, FILE_NAME, __LINE__);
   }
-  DIR* dirp = env->get_pointer(env, stack, obj_dirp);
+  MY_DIR* dirp = env->get_pointer(env, stack, obj_dirp);
   
+#if defined(_WIN32)
+  int32_t status = _wclosedir(dirp);
+#else
   int32_t status = closedir(dirp);
+#endif
+
   if (status == -1) {
     env->die(env, stack, "[System Error]closedir() failed:%s.", env->strerror_nolen(env, stack, errno), __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
@@ -1414,10 +1438,15 @@ int32_t SPVM__Sys__IO__readdir(SPVM_ENV* env, SPVM_VALUE* stack) {
   if (!obj_dirp) {
     return env->die(env, stack, "The directory stream $dirp must be defined.", __func__, FILE_NAME, __LINE__);
   }
-  DIR* dirp = env->get_pointer(env, stack, obj_dirp);
+  MY_DIR* dirp = env->get_pointer(env, stack, obj_dirp);
   
   errno = 0;
-  struct dirent* dirent = readdir(dirp);
+#if defined(_WIN32)
+  MY_DIRENT* dirent = _wreaddir(dirp);
+#else
+  MY_DIRENT* dirent = readdir(dirp);
+#endif
+
   if (errno != 0) {
     env->die(env, stack, "[System Error]readdir() failed:%s.", env->strerror_nolen(env, stack, errno), __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
@@ -1443,10 +1472,14 @@ int32_t SPVM__Sys__IO__rewinddir(SPVM_ENV* env, SPVM_VALUE* stack) {
     return env->die(env, stack, "The directory stream $dirp must be defined.", __func__, FILE_NAME, __LINE__);
   }
   
-  DIR* dirp = env->get_pointer(env, stack, obj_dirp);
+  MY_DIR* dirp = env->get_pointer(env, stack, obj_dirp);
   
+#if defined(_WIN32)
+  _wrewinddir(dirp);
+#else
   rewinddir(dirp);
-  
+#endif
+
   return 0;
 }
 
@@ -1456,9 +1489,14 @@ int32_t SPVM__Sys__IO__telldir(SPVM_ENV* env, SPVM_VALUE* stack) {
   if (!obj_dirp) {
     return env->die(env, stack, "The directory stream $dirp must be defined.", __func__, FILE_NAME, __LINE__);
   }
-  DIR* dirp = env->get_pointer(env, stack, obj_dirp);
+  MY_DIR* dirp = env->get_pointer(env, stack, obj_dirp);
   
+#if defined(_WIN32)
+  int64_t offset = _wtelldir(dirp);
+#else
   int64_t offset = telldir(dirp);
+#endif
+
   if (offset == -1) {
     env->die(env, stack, "[System Error]telldir() failed:%s.", env->strerror_nolen(env, stack, errno), __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
@@ -1479,14 +1517,18 @@ int32_t SPVM__Sys__IO__seekdir(SPVM_ENV* env, SPVM_VALUE* stack) {
     return env->die(env, stack, "The directory stream $dirp must be defined.", __func__, FILE_NAME, __LINE__);
   }
   
-  DIR* dirp = env->get_pointer(env, stack, obj_dirp);
+  MY_DIR* dirp = env->get_pointer(env, stack, obj_dirp);
   
   if (!(offset >= 0)) {
     return env->die(env, stack, "The offset $offset must be less than or equal to 0.", __func__, FILE_NAME, __LINE__);
   }
   
+#if defined(_WIN32)
+  _wseekdir(dirp, offset);
+#else
   seekdir(dirp, offset);
-  
+#endif
+
   return 0;
 }
 
