@@ -11,80 +11,6 @@ static const char* FILE_NAME = "Sys/IO/Windows.c";
 
 #include "Sys-Windows.h"
 
-// Same as Perl's win32_unlink in Win32.c, but call APIs for wide characters(W instead of A and _w prefixed).
-static int
-win32_unlink(const wchar_t *filename)
-{
-  int ret;
-  DWORD attrs;
-  
-  attrs = GetFileAttributesW(filename);
-  if (attrs == 0xFFFFFFFF) {
-    errno = ENOENT;
-    return -1;
-  }
-  
-  if (attrs & FILE_ATTRIBUTE_READONLY) {
-    (void)SetFileAttributesW(filename, attrs & ~FILE_ATTRIBUTE_READONLY);
-    ret = _wunlink(filename);
-    if (ret == -1)
-        (void)SetFileAttributesW(filename, attrs);
-  }
-  else if ((attrs & (FILE_ATTRIBUTE_REPARSE_POINT | FILE_ATTRIBUTE_DIRECTORY))
-    == (FILE_ATTRIBUTE_REPARSE_POINT | FILE_ATTRIBUTE_DIRECTORY)
-         && is_symlink_name(filename)) {
-    ret = _wrmdir(filename);
-  }
-  else {
-    ret = _wunlink(filename);
-  }
-  
-  return ret;
-}
-
-// Same as Perl's win32_rename in Win32.c, but call APIs for wide characters(W instead of A and _w prefixed).
-static int
-win32_rename(const wchar_t *oname, const wchar_t *newname)
-{
-    wchar_t szOldName[MAX_PATH+1];
-    BOOL bResult;
-    DWORD dwFlags = MOVEFILE_COPY_ALLOWED;
-    dTHX;
-
-    if (_wcsicmp(newname, oname))
-        dwFlags |= MOVEFILE_REPLACE_EXISTING;
-    wcscpy(szOldName, oname);
-    
-    bResult = MoveFileExW(szOldName,newname, dwFlags);
-    
-    if (!bResult) {
-        DWORD err = GetLastError();
-        switch (err) {
-        case ERROR_BAD_NET_NAME:
-        case ERROR_BAD_NETPATH:
-        case ERROR_BAD_PATHNAME:
-        case ERROR_FILE_NOT_FOUND:
-        case ERROR_FILENAME_EXCED_RANGE:
-        case ERROR_INVALID_DRIVE:
-        case ERROR_NO_MORE_FILES:
-        case ERROR_PATH_NOT_FOUND:
-            errno = ENOENT;
-            break;
-        case ERROR_DISK_FULL:
-            errno = ENOSPC;
-            break;
-        case ERROR_NOT_ENOUGH_QUOTA:
-            errno = EDQUOT;
-            break;
-        default:
-            errno = EACCES;
-            break;
-        }
-        return -1;
-    }
-    return 0;
-}
-
 static int
 win32_symlink(SPVM_ENV* env, SPVM_VALUE* stack, const wchar_t *oldfile, const wchar_t *newfile)
 {
@@ -194,6 +120,7 @@ win32_symlink(SPVM_ENV* env, SPVM_VALUE* stack, const wchar_t *oldfile, const wc
 
 #endif // _WIN32
 
+// This logic is the same as Perl's win32_unlink in Win32.c, and UTF-8 arguments are supported.
 int32_t SPVM__Sys__IO__Windows__unlink(SPVM_ENV* env, SPVM_VALUE* stack) {
 #if !defined(_WIN32)
   env->die(env, stack, "Sys::IO::Windows#unlink method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
@@ -202,30 +129,60 @@ int32_t SPVM__Sys__IO__Windows__unlink(SPVM_ENV* env, SPVM_VALUE* stack) {
   
   int32_t error_id = 0;
   
-  void* obj_pathname = stack[0].oval;
-  if (!obj_pathname) {
-    return env->die(env, stack, "The path $pathname must be defined.", __func__, FILE_NAME, __LINE__);
+  void* obj_path = stack[0].oval;
+  
+  if (!obj_path) {
+    return env->die(env, stack, "The path $path must be defined.", __func__, FILE_NAME, __LINE__);
   }
   
-  const char* pathname = env->get_chars(env, stack, obj_pathname);
+  const char* path = env->get_chars(env, stack, obj_path);
   
-  wchar_t* pathname_w = utf8_to_win_wchar(env, stack, pathname, &error_id, __func__, FILE_NAME, __LINE__);
+  wchar_t* path_w = utf8_to_win_wchar(env, stack, path, &error_id, __func__, FILE_NAME, __LINE__);
   if (error_id) {
     return error_id;
   }
   
-  int32_t status = win32_unlink(pathname_w);
+  int32_t status = -1;
+  
+  DWORD attrs = GetFileAttributesW(path_w);
+  
+  if (attrs == 0xFFFFFFFF) {
+    errno = ENOENT;
+    status = -1;
+    goto END_OF_FUNC;
+  }
+  
+  if (attrs & FILE_ATTRIBUTE_READONLY) {
+    SetFileAttributesW(path_w, attrs & ~FILE_ATTRIBUTE_READONLY);
+    status = _wunlink(path_w);
+    if (status == -1) {
+      SetFileAttributesW(path_w, attrs);
+    }
+  }
+  else if ((attrs & (FILE_ATTRIBUTE_REPARSE_POINT | FILE_ATTRIBUTE_DIRECTORY))
+    == (FILE_ATTRIBUTE_REPARSE_POINT | FILE_ATTRIBUTE_DIRECTORY)
+         && is_symlink_name(path_w))
+  {
+    status = _wrmdir(path_w);
+  }
+  else {
+    status = _wunlink(path_w);
+  }
+  
+  END_OF_FUNC:
   
-  stack[0].ival = status;
   if (status == -1) {
-    env->die(env, stack, "[System Error]unlink() failed:%s. $pathname is \"%s\".", env->strerror_nolen(env, stack, errno), pathname, __func__, FILE_NAME, __LINE__);
+    env->die(env, stack, "[System Error]unlink() failed(%d: %s). $path=\"%s\".", errno, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
   }
   
+  stack[0].ival = status;
+  
   return 0;
 #endif
 }
 
+// This logic is the same as Perl's win32_rename in Win32.c, and UTF-8 arguments are supported.
 int32_t SPVM__Sys__IO__Windows__rename(SPVM_ENV* env, SPVM_VALUE* stack) {
 #if !defined(_WIN32)
   env->die(env, stack, "Sys::IO::Windows#rename method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
@@ -256,10 +213,49 @@ int32_t SPVM__Sys__IO__Windows__rename(SPVM_ENV* env, SPVM_VALUE* stack) {
     return error_id;
   }
   
-  int32_t status = win32_rename(oldpath_w, newpath_w);
+  DWORD flags = MOVEFILE_COPY_ALLOWED;
+  if (!(_wcsicmp(newpath_w, oldpath_w) == 0)) {
+    flags |= MOVEFILE_REPLACE_EXISTING;
+  }
+  
+  int32_t success = MoveFileExW(oldpath_w, newpath_w, flags);
+  
+  int32_t status = -1;
+  if (success) {
+    status = 0;
+  }
+  else {
+    DWORD last_error = GetLastError();
+    switch (last_error) {
+      case ERROR_BAD_NET_NAME:
+      case ERROR_BAD_NETPATH:
+      case ERROR_BAD_PATHNAME:
+      case ERROR_FILE_NOT_FOUND:
+      case ERROR_FILENAME_EXCED_RANGE:
+      case ERROR_INVALID_DRIVE:
+      case ERROR_NO_MORE_FILES:
+      case ERROR_PATH_NOT_FOUND:
+      {
+        errno = ENOENT;
+        break;
+      }
+      case ERROR_DISK_FULL: {
+        errno = ENOSPC;
+        break;
+      }
+      case ERROR_NOT_ENOUGH_QUOTA: {
+        errno = EDQUOT;
+        break;
+      }
+      default: {
+        errno = EACCES;
+        break;
+      }
+    }
+  }
   
   if (status == -1) {
-    env->die(env, stack, "[System Error]rename() failed:%s. $oldpath is \"%s\". $newpath is \"%s\".", env->strerror_nolen(env, stack, errno), oldpath, newpath, __func__, FILE_NAME, __LINE__);
+    env->die(env, stack, "[System Error]MoveFileExW() for renaming failed(%d: %s). $oldpath=\"%s\", $newpath=\"%s\".", errno, env->strerror_nolen(env, stack, errno), oldpath, newpath, __func__, FILE_NAME, __LINE__);
     return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
   }
   
@@ -274,6 +270,7 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
   env->die(env, stack, "Sys::IO::Windows#win_readlink method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
   return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_NOT_SUPPORTED_CLASS;
 #else
+
   int32_t error_id = 0;
   void* obj_path = stack[0].oval;
   if (!obj_path) {
@@ -288,36 +285,47 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
   DWORD fileattr = GetFileAttributesW(path_w);
   if (fileattr == INVALID_FILE_ATTRIBUTES) {
     translate_to_errno();
+    env->die(env, stack, "[System Error]GetFileAttributesW() failed(%d: %s). $path=\"%s\".", errno, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);    
     error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
     goto END_OF_FUNC;
   }
   
-  if (!(fileattr & FILE_ATTRIBUTE_REPARSE_POINT)) {
-    /* not a symbolic link */
-    errno = EINVAL;
-    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
-    goto END_OF_FUNC;
-  }
-  
-  HANDLE hlink =
-    CreateFileW(path_w, GENERIC_READ, 0, NULL, OPEN_EXISTING,
-                FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
-  if (hlink == INVALID_HANDLE_VALUE) {
-    translate_to_errno();
-    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
-    goto END_OF_FUNC;
-  }
-  
-  
   MY_REPARSE_DATA_BUFFER linkdata;
   DWORD linkdata_returned;
-  
-  if (!DeviceIoControl(hlink, FSCTL_GET_REPARSE_POINT, NULL, 0, &linkdata, sizeof(linkdata), &linkdata_returned, NULL)) {
-    translate_to_errno();
-    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
-    goto END_OF_FUNC;
+  HANDLE handle = NULL;
+  if (fileattr & FILE_ATTRIBUTE_REPARSE_POINT) {
+    handle = CreateFileW_reparse_point_for_read(path_w);
+    
+    if (handle == INVALID_HANDLE_VALUE) {
+      translate_to_errno();
+      env->die(env, stack, "[System Error]CreateFileW() failed when opening a file(%d: %s). $path=\"%s\".", errno, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+      error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+      goto END_OF_FUNC;
+    }
+    
+    if (!DeviceIoControl(handle, FSCTL_GET_REPARSE_POINT, NULL, 0, &linkdata, sizeof(linkdata), &linkdata_returned, NULL)) {
+      translate_to_errno();
+      env->die(env, stack, "[System Error]DeviceIoControl() failed(%d: %s). $path=\"%s\".", errno, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+      error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+      goto END_OF_FUNC;
+    }
+  }
+  else {
+    handle = CreateFileW_for_read(path_w);
+    
+    if (handle == INVALID_HANDLE_VALUE) {
+      translate_to_errno();
+      env->die(env, stack, "[System Error]CreateFileW() failed when opening a file(%d: %s). $path=\"%s\".", errno, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+      error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+      goto END_OF_FUNC;
+    }
+    else {
+      errno = EINVAL;
+      env->die(env, stack, "[System Error]This file is not a reparse point. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
+      error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+      goto END_OF_FUNC;
+    }
   }
-  
   
   const wchar_t* PathBuffer = NULL;
   int32_t PrintNameOffset = -1;
@@ -327,7 +335,8 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
       const MY_SYMLINK_REPARSE_BUFFER * const sd =
         &linkdata.Data.SymbolicLinkReparseBuffer;
       if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.SymbolicLinkReparseBuffer.PathBuffer)) {
-        errno = EINVAL;
+        errno = ENOMEM;
+        env->die(env, stack, "[System Error]The data DeviceIoControl() retruned is invalid. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
         error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
         goto END_OF_FUNC;
       }
@@ -342,7 +351,8 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
       const MY_MOUNT_POINT_REPARSE_BUFFER * const rd =
         &linkdata.Data.MountPointReparseBuffer;
       if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.MountPointReparseBuffer.PathBuffer)) {
-        errno = EINVAL;
+        errno = ENOMEM;
+        env->die(env, stack, "[System Error]The data DeviceIoControl() retruned is invalid. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
         error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
         goto END_OF_FUNC;
       }
@@ -354,6 +364,7 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
     }
     default: {
       errno = EINVAL;
+      env->die(env, stack, "[System Error]The type of the reparse point must be IO_REPARSE_TAG_SYMLINK or IO_REPARSE_TAG_MOUNT_POINT. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
       error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
       goto END_OF_FUNC;
     }
@@ -366,8 +377,8 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
                         NULL, 0, NULL, NULL);
                         
   if (bytes_out == 0) {
-    /*() failed.conversion from unicode to ANSI or otherwise() failed.*/
-    errno = EINVAL;
+    env->die(env, stack, "[System Error]WideCharToMultiByte() failed. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    errno = EILSEQ;
     error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
     goto END_OF_FUNC;
   }
@@ -381,23 +392,19 @@ int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
                         PrintNameLength/2,
                         link_text, bytes_out, NULL, NULL);
   if (bytes_out == 0) {
-    /*() failed.conversion from unicode to ANSI or otherwise() failed.*/
-    errno = EINVAL;
+    env->die(env, stack, "[System Error]WideCharToMultiByte() failed. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    errno = EILSEQ;
     error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
     goto END_OF_FUNC;
   }
   
   END_OF_FUNC:
   
-  if (!(hlink == INVALID_HANDLE_VALUE)) {
-    CloseHandle(hlink);
+  if (!(handle == INVALID_HANDLE_VALUE)) {
+    CloseHandle(handle);
   }
   
   if (error_id) {
-    if (errno) {
-      env->die(env, stack, "[System Error]win_readlink() failed:%s. $path is \"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
-    }
-    
     return error_id;
   }
   
@@ -490,9 +497,7 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
     return error_id;
   }
   
-  HANDLE handle = CreateFileW(resolved_link_text_w, FILE_READ_ATTRIBUTES,
-                    FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE,
-                    NULL, OPEN_EXISTING, FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, NULL);
+  HANDLE handle = CreateFileW_reparse_point_for_read(resolved_link_text_w);
   
   if (handle == INVALID_HANDLE_VALUE) {
     translate_to_errno();
@@ -503,7 +508,7 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
   int32_t needed_len = GetFinalPathNameByHandleW(handle, NULL, 0, 0);
   
   if (needed_len == 0) {
-    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path:\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
     error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
     goto END_OF_FUNC;
   }
@@ -514,7 +519,7 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
   int32_t len = GetFinalPathNameByHandleW(handle, resolved_path_w, needed_len, 0);
   
   if (len == 0) {
-    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path:\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path=\"%s\".", path, __func__, FILE_NAME, __LINE__);
     error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
     goto END_OF_FUNC;
   }
@@ -556,7 +561,7 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
   
   if (error_id) {
     if (errno) {
-      env->die(env, stack, "[System Error]win_realpath() failed:%s. $path:\"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+      env->die(env, stack, "[System Error]win_realpath() failed:%s. $path=\"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
     }
     
     return error_id;
