@@ -8,7 +8,7 @@ SPVM::Sys::Time::Tms - struct tms in the C language
 
 =head1 Description
 
-The Sys::Time::Tms class in L<SPVM> represents L<struct tms|https://linux.die.net/man/2/times> in the C language.
+Sys::Time::Tms class in L<SPVM> represents L<struct tms|https://linux.die.net/man/2/times> in the C language.
 
 =head1 Usage
   
