@@ -403,7 +403,7 @@ is_symlink(HANDLE h) {
 // Same as Perl's is_symlink_name in Win32.c, but call APIs for wide characters(W instead of A).
 static BOOL
 is_symlink_name(const wchar_t *name) {
-    HANDLE f = CreateFileW(name, GENERIC_READ, 0, NULL, OPEN_EXISTING,
+    HANDLE f = CreateFileW(name, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING,
                            FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
     BOOL result;
 
@@ -600,6 +600,25 @@ translate_ft_to_time_t(FILETIME ft) {
     return retval;
 }
 
+static HANDLE CreateFileW_for_read_common(wchar_t* path_w, int32_t file_flag) {
+
+  HANDLE handle = CreateFileW(path_w, GENERIC_READ,
+    FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING,
+    file_flag|FILE_FLAG_BACKUP_SEMANTICS, 0
+  );
+  
+  return handle;
+}
+static HANDLE CreateFileW_for_read(wchar_t* path_w) {
+  
+  return CreateFileW_for_read_common(path_w, 0);
+}
+
+static HANDLE CreateFileW_reparse_point_for_read(wchar_t* path_w) {
+
+  return CreateFileW_for_read_common(path_w, FILE_FLAG_OPEN_REPARSE_POINT);
+}
+
 #endif // defined(_WIN32)
 
 #endif // SPVM__SYS__WINDOWS__H
